<?php
/**
 * ffmpeg Webcam management function
 *
 * Included by viewFunctions.php
 * parameters :
 * - $completeFilename : complete filename of image to view
 * - $command : command
 * - $extraData : [image outer width]x[image outer height]
 *
 * Main resource config parameters used :
 * "reencodedVideo": true to allow flv real-time or pre-encoding
 * "flashStream":  true for real-time encoding, false for preencoding *** not used anymore since V210Q as preencoding is no more allowed => set to true
 * "flashOutputSize": flash (pre or real-time) encoding width (160, 320, 400, 640, 900, 1280)
 * "flashOutputQuality": flash (pre or real-time) encoding quality ('original' | 'high' | 'medium' | 'low')
 *
 * "allowOriginalVideo": allow direct viewing of source video (with any allowed player)
 * "useQTVideo": true to allow qt/mp4 preencoding when compatible device found
 * "qtOutputSize": qt/mp4 preencoding width (default value set to 480 = iPhone width)
 * "qtOutputQuality": qt/mp4 preencoding quality ('high' | 'medium' | 'low')

 *
 *
 * PHP version 5
 *
 * LICENSE: This source file is subject to version 3.0 of the PHP license
 * that is available through the world-wide-web at the following URI:
 * http://www.php.net/license/3_0.txt.  If you did not receive a copy of
 * the PHP License and are unable to obtain it through the web, please
 * send a note to license@php.net so we can mail you a copy immediately.
 *
 * @category   NA
 * @package    NA
 * @author     Nicolas Bruley / Peer 2 World <contact@weezo.net>
 * @copyright  2005-2009 Nicolas Bruley / Peer 2 World
 * @license    http://www.php.net/license/3_0.txt  PHP License 3.0
 * @version    CVS: $Id:$
 * @link       http://www.weezo.net
 * @since      File available since Release 1.1.0
 */

/**
 * @desc Return a list of video and audio sources
 *
 * @return array('video'=>array(id=>array('label'=>,'resolution'=>)...))
 */
function webcamGetList($useCache=true){
	$webcams=array(
		'video'=>array(),
		'audio'=>array()
	);

	// To limit ffmpeg access, refresh webcam list every 5s only
	if(!$useCache && (cfMGetVar('webcams') && !isset($_GET['webcamUpdateList']) && !isset($_POST['webcamUpdateList']) && (time()-cfMGetVar('webcamListUpdateTime')<5))) return cfMGetVar('webcams');

	/**
	 * Windows
	 */
	if(WEnv::host()->isWindows()){
		$cl='"'.cfFfmpegExe().'" -list_devices true -f dshow -i dummy';
		cfFfmpegLog($cl,'Get webcams list');
		$list=cfLinesFromText(cfStreamProc($cl,array('stderr'=>'return')));

		if(is_array($list)) foreach ($list as $line){
			$line=trim($line);
			if(strpos($line,'DirectShow video devices')){
				$inVideo=1;
				$inAudio=0;
				continue;
			}
			if(strpos($line,'DirectShow audio devices')){
				$inAudio=1;
				$inVideo=0;
				continue;
			}
			if(cfCmpLeft($line,'[dshow') && strpos($line,'"')){
				$name=substr($line,strpos($line,'"')+1);
				$name=substr($name,0,strlen($name)-1);
				if($inAudio) $webcams['audio'][md5($name)]=array(
					'label'=>$name,
					'capabilities'=>webcamAcquireAudioCapabilities($name)
				);
				if($inVideo) {
					$webcams['video'][md5($name)]=array(
						'label'=>$name,
						'resolution'=>webcamAcquireResolution($name,$useCache)
					);
				}
			}
			else{
				$inAudio=0;
				$inVideo=0;
			}
		}
	}

	/**
	 * Linux
	 */
	if(WEnv::host()->isLinux()){
		$i=0;

		// Video src
		while (file_exists('/dev/video'.$i)){
			$name='/dev/video'.$i;
			$webcams['video'][md5($name)]=array(
				'label'=>$name,
				'resolution'=>webcamAcquireResolution($name,$useCache)
			);
			$i++;
		}

		// Audio src
		while (file_exists('/dev/dsp'.$i)){
			$name='/dev/dsp'.$i;
			$webcams['video'][md5($name)]=array(
				'label'=>$name,
				'capabilities'=>false
			);
			$i++;
		}
	}


	return $webcams;
}

/**
 * @desc Get webcam video resolution with ffmpeg
 *
 * @param string $label
 * @return string width+"x"+height
 */
function webcamAcquireResolution($label,$useCache=true){
	// In order not to start webcam every time
	$knownDevices=@unserialize(cfGGetVar('webcamResolutions'));
	if($useCache && isset($knownDevices[$label])) return $knownDevices[$label];

	$width=0; $height=0;

	// Windows: use list_option parameter with ffmpeg to get full detail
	if(WEnv::host()->isWindows()){
		$cl='"'.cfFfmpegExe().'" -list_options true -f dshow -i video="'.cfUTF8Decode($label).'"';
		cfFfmpegLog($cl,'Get webcam video capabilities');

		$list=cfLinesFromText(cfStreamProc($cl,array('stderr'=>'return')));

		// Parse ffmpeg output
		if(is_array($list)) foreach ($list as $line){
			$line=trim($line);
			if(strpos($line,'max s=')){
				$line=substr($line,strpos($line,'max s=')+6);
				$line=substr($line,0, strpos($line,' '));
				list($w,$h)=@explode('x',$line);
				if(is_numeric($w) && is_numeric($h)){
					if($w>$width){
						$width=$w;
						$height=$h;
					}
				}
			}
		}

		// Store result
		$knownDevices[$label]=$width.'x'.$height;
		cfGUpdateVar('webcamResolutions',serialize($knownDevices));

		//if($width==0) return false;
	}

	// Linux: take a screenshot with ffmpeg
	elseif(WEnv::host()->isLinux()){
		$file=cfAppTempDir().'/webcam_tmp.jpg';
		@unlink($file);
		$cl='"'.cfFfmpegExe().'" -f video4linux2 -qscale 1 -i '.$label.' -vframes 1 '.$file;
		@list($width,$height)=getimagesize($file);
		@unlink($file);
	}

	if(!$width) $width=640;
	if(!$height) $height=480;

	return $width.'x'.$height;
}

/**
 * Parse args in ffmpeg ouput
 *
 * @param string $string
 * @param string $needle
 * @return string
 */
function webcamStrGetArg($string,$needle){
	if(($p=strpos($string,$needle.'='))===false) return false;
	$string=trim(substr($string,$p+strlen($needle)+1));
	if($p=strpos($string,' ')) $string=substr($string,0,$p);
	return $string;
}

/**
 * @desc Get webcam audio capabilities with ffmpeg
 *
 * @param string $label
 * @return array
 */
function webcamAcquireAudioCapabilities($label,$useCache=true){
	// In order not to start webcam every time
	$knownDevices=@unserialize(cfGGetVar('webcamAudioCapabilities'));
	if($useCache && isset($knownDevices[$label])) return $knownDevices[$label];

	$cl='"'.cfFfmpegExe().'" -list_options true -f dshow -i audio="'.cfUTF8Decode($label).'"';
	$list=cfLinesFromText(cfStreamProc($cl,array('stderr'=>'return')));

	$caps=false;

	// Parse ffmpeg output
	if(is_array($list)) foreach ($list as $line){
		$line=trim($line);
		if(strpos($line,'max') && strpos($line,'min') && strpos($line,'ch=') && strpos($line,'rate=') ){
			$caps=array(
				'min'=>array(),
				'max'=>array(),
			);
			$min=substr($line,strpos($line,'min '));
			$max=substr($min,strpos($min,'max '));
			$min=substr($line,0,strpos($line,'max '));

			$caps['min']['ch']=webcamStrGetArg($min,'ch');
			$caps['min']['bits']=webcamStrGetArg($min,'bits');
			$caps['min']['rate']=webcamStrGetArg($min,'rate');

			$caps['max']['ch']=webcamStrGetArg($max,'ch');
			$caps['max']['bits']=webcamStrGetArg($max,'bits');
			$caps['max']['rate']=webcamStrGetArg($max,'rate');
			break;
		}
	}

	// Store result
	$knownDevices[$label]=$caps;
	cfGSetVar('webcamAudioCapabilities',$knownDevices);

	return $caps;
}

/**
 * @desc Return width and height resolution of a webcam
 *
 * @param string $webcamId: id of webcam
 * @return array: (width,height)
 */
function webcamGetResolution($webcamId){
	$webcams=webcamGetList();
	$resolution=@$webcams['video'][$webcamId]['resolution'];
	if(strpos($resolution,'x')) return explode('x',$resolution,2);
	return array(0,0);
}

/**
 * @desc Function called by /ext/video.php function to check requested file access rights
 *
 * @param string_type $passedFilename
 * @return mixed
 */
function webcamCheckAccessRights($passedFilename){
	@list($w,$videoId,$audioId)=@explode(':',$passedFilename);
	if($w!=='webcam') return false;

	if($videoId!=cfRGetVar('videoSource')) return false;
	if($audioId && $audioId!=cfRGetVar('audioSource')) return false;

	return $passedFilename;
}

/**
 * @desc Return part of ffmpeg command line that must be set after "-i" parameter to stream webcam
 *
 * @param unknown_type $passedFilename
 * @return unknown
 */
function webcamFfmpegCommandLineInput($passedFilename,&$soundMinQuality,&$soundMaxQuality){
	@list($foo,$videoId,$audioId)=@explode(':',$passedFilename);

	$webcams=webcamGetList();

	$videoDeviceName=cfUTF8Decode(@$webcams['video'][$videoId]['label']);
	$audioDeviceName=cfUTF8Decode(@$webcams['audio'][$audioId]['label']);

	$cl=' video="'.$videoDeviceName.'"';

	// Audio parameters
	if($audioDeviceName) {
		$cl.=':audio="'.$audioDeviceName.'"';

		$caps=webcamAcquireAudioCapabilities(@$webcams['audio'][$audioId]['label']);
		$soundMinQuality=' -af aresample -ar '.$caps['min']['rate'].' -ac '.(($caps['min']['ch'])?$caps['min']['ch']:1);
		$soundMaxQuality=' -af aresample -ar '.$caps['max']['rate'].' -ac '.(($caps['max']['ch'])?$caps['max']['ch']:1);
	}

	$cl.=' -r 15 -f dshow ';

	return $cl;
}


/**
 * @desc Take a screenshot of webcam
 *
 * @param string $webcamId
 * @param integer $width
 * @return false on failure
 */
function webcamScreenshot($videoId,$width=false){
	$webcams=webcamGetList();

	$videoDeviceName=cfUTF8Decode(@$webcams['video'][$videoId]['label']);
	list($w,$h)=@explode('x',@$webcams['video'][$videoId]['resolution']);
	if(!$w || !$h) return false;
	if(!$width) $width=$w;

	list($width,$height)=cfThumbnailDimensions($w,$h,$width);

	// Get video dimensions by getting dimensions of a generated thumbnail... classe... pas classe :)
	$cl=' -f dshow -i video="'.$videoDeviceName.'" -vframes 1 -an -s '.$width.'x'.$height.' -f image2 -';

	WHeaders::contentType('jpg');
	WHeaders::noCache();

	// Generate image from video
	cfFfmpegLog($cl,'Get '.$videoDeviceName.' webcam screenshot');
	cfStreamProc('"'.cfFfmpegExe().'"'.$cl,array('stdout'=>'stdout'),array('speedLimit'=>WEnv::user()->getVar('downloadSpeedLimit')));
	exit;
}